from sklearn.preprocessing import MultiLabelBinarizer
import evaluate
import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import classification_report, confusion_matrix
from collections import Counter

class Categorizer:
    def __init__(self, n, threshold, model_path, output_path, classifier_model):
        """
        Initializes the Categorizer class.

        Parameters:
        n (int): Number of maximum labels for a single mention.
        threshold (float): Threshold for classification of a sample.
        model_path (str): Path to the pretrained model.
        output_path (str): Path to save the output.
        classifier_model (str): Name of the classifier model.
        """
        self.n = n
        self.threshold = threshold
        self.labels = ['ACTIVIDAD', 'COMUNIDAD', 'DEPARTAMENTO', 'ENFERMEDAD', 'FAC_GEN', 'FAC_NOM', 'FARMACO', 'GEO_GEN', 'GEO_NOM', 'GPE_GEN', 'GPE_NOM', 'HUMAN', 'IDIOMA', 'MORFOLOGIA_NEOPLASIA', 'NO_CATEGORY', 'PROCEDIMIENTO', 'PROFESION', 'SINTOMA', 'SITUACION_LABORAL', 'SPECIES', 'TRANSPORTE']
        self.output_path = output_path
        self.classifier_model = classifier_model
        self.model_path = model_path
        self.model = self.initialize_pretrained_model()

    def evaluate_model(self, y_pred, y_test, mcm, classification_report):
        """
        Evaluates the model's performance and calls the selected evauation functions.

        Parameters:
        y_pred (list): Predicted labels. Labels are provided in the shape of vectors.
        y_test (list): True labels. Labels are provided in the shape of vectors.
        mcm (bool): Whether to create a multi-label confusion matrix heatmap.
        classification_report (bool): Whether to display a classification report.

        Returns:
        dict: Dictionary containing computed evaluation metrics.
        """
        if classification_report:
            self.classification_report(y_pred, y_test)
        if mcm:
            self.mcm_heatmap(y_pred, y_test)
        return self.compute_metrics(y_pred, y_test)

    def lambda_evaluate_model(self, mcm, classification_report):
        """
        Returns a lambda function for evaluating the model.

        Parameters:
        mcm (bool): Whether to create a multi-label confusion matrix heatmap.
        classification_report (bool): Whether to display a classification report.

        Returns:
        function: Lambda function for model evaluation.
        """
        return lambda y_pred, y_test: self.evaluate_model(y_pred, y_test, mcm, classification_report)

    def compute_metrics(self, y_pred, y_test):
        """
        Computes evaluation metrics: F-score, accuracy, and number of samples within each class with no given label.

        Parameters:
        y_pred (list): Predicted labels. Labels are provided in the shape of vectors.
        y_test (list): True labels. Labels are provided in the shape of vectors.

        Returns:
        dict: Dictionary containing computed evaluation metrics.
        """
        multilabel_f1_metric = evaluate.load("f1", "multilabel")
        multilabel_accuracy_metric = evaluate.load("accuracy", "multilabel")
        f1 = multilabel_f1_metric.compute(predictions=y_pred, references=y_test, average="micro")["f1"]
        accuracy = multilabel_accuracy_metric.compute(predictions=y_pred, references=y_test)["accuracy"]

        y_pred = np.array(y_pred)
        y_test = np.array(y_test)

        no_label_samples = []
        for idx, pred in enumerate(y_pred):
            if np.all(pred == 0):
                true_labels = [self.labels[i] for i, value in enumerate(y_test[idx]) if value == 1]
                no_label_samples.extend(true_labels)

        label_counts = Counter(no_label_samples)
        label_counts_dict = dict(label_counts)
        return {"f1": f1, "accuracy": accuracy, "Classes with no given label": label_counts_dict}

    def classification_report(self, y_pred, y_test):
        """
        Generates and displays a classification report.

        Parameters:
        y_pred (list): Predicted labels. Labels are provided in the shape of vectors.
        y_test (list): True labels. Labels are provided in the shape of vectors.
        """
        y_pred = np.array(y_pred)
        y_test = np.array(y_test)

        print(classification_report(y_test, y_pred, target_names=self.labels))

    def mcm_heatmap(self, y_pred, y_test):
        """
        Generates and displays a multi-label confusion matrix heatmap.

        Parameters:
        y_pred (list): Predicted labels. Labels are provided in the shape of vectors.
        y_test (list): True labels. Labels are provided in the shape of vectors.
        """
        y_pred = np.array(y_pred)
        y_test = np.array(y_test)

        samples_with_predictions = np.any(y_pred, axis=1)

        y_pred_with_predictions = y_pred[samples_with_predictions]
        y_test_with_predictions = y_test[samples_with_predictions]

        confusion_matrix_multi = confusion_matrix(y_test_with_predictions.argmax(axis=1), y_pred_with_predictions.argmax(axis=1), labels=np.arange(len(self.labels)))
    
        plt.figure(figsize=(10, 7))
        plt.imshow(confusion_matrix_multi, interpolation="nearest", cmap=plt.cm.Blues)
        plt.title("Multi-Label Confusion Matrix")
        plt.colorbar()

        tick_marks = np.arange(len(self.labels))
        plt.xticks(tick_marks, self.labels, rotation=45)
        plt.yticks(tick_marks, self.labels)

        for i in range(len(self.labels)):
            for j in range(len(self.labels)):
                plt.text(j, i, str(confusion_matrix_multi[i, j]), ha="center", va="center", color="white" if confusion_matrix_multi[i, j] > confusion_matrix_multi.max() / 2 else "black")

        plt.tight_layout()
        plt.ylabel("True Labels")
        plt.xlabel("Predicted Labels")
        plt.show()



import pandas as pd
import time
import sys, os, re

# This script can be used to test the TermExtractor pipeline using any desired extraction and categorization method on the given sample files
# Note that the script also returns the computed execution time and saves the results

# Arguments:
# 1 - text files folder path: '../../data/text_files'
# 2 - extraction_method: 'textrank'
# 3 - categorization_method: 'setfit'
# 5 - output_dir path: "./extracted_keywords.tsv"

#set the path to the library
general_path = os.getcwd().split("KeyCARE")[0]+"KeyCARE/"
sys.path.append(general_path+'src/keycare/')
from ..TermExtractor import TermExtractor

start_time = time.time()

keywords = []
extractor = TermExtractor(extraction_method=[sys.argv[2]], categorization_method=sys.argv[3])

docs = os.listdir(sys.argv[1])
for doc in docs:
    if doc.endswith(".txt"):
        with open(os.path.join(sys.argv[1], doc), "r") as file:
            text = file.read()
        extractor(text)
        kws = extractor.keywords
        kws_with_file = [(doc[:-4], kw.span, kw.text, kw.score, kw.label, kw.extraction_method, kw.categorization_method) for kw in kws]
        keywords.extend(kws_with_file)

df = pd.DataFrame(keywords, columns=["file", "span", "text", "score", "label", "extraction_method", "categorization_method"])
df.to_csv(sys.argv[4], sep="\t")

end_time = time.time()
execution_time = end_time - start_time
print(f"Execution time: {execution_time} seconds")


from setfit import SetFitModel
from .Relator import Relator

class SetFitRelator(Relator):
    def __init__(self, n, threshold, model_path):
        """
        Initializes SetFitRelator, a class inheriting from Relator.

        Parameters:
        n (int): Maximum number of labels for a single relation.
        threshold (float): Threshold value for mentions relation.
        model_path (str): Path to the model.
        """
        super().__init__(n, threshold, model_path)
    
    def initialize_pretrained_model(self, model_path):
        """
        Initializes a pretrained SetFitModel based on the provided model_path.

        Parameters:
        model_path (str): Path to the pretrained model.

        Returns:
        object: SetFitModel instance.
        """
        if model_path is None:
            path = 'BSC-NLP4BIA/biomedical-semantic-relation-classifier-setfit'
            model = SetFitModel.from_pretrained(path)
        else:
            model = SetFitModel.from_pretrained(model_path)
        return model

    def compute_relation(self, source, target):
        """
        Computes relations between source and target entities.

        Parameters:
        source (list): List of source entities.
        target (list): List of target entities.

        Returns:
        list: List of labels representing computed relations.
        """
        final_labels = list()
        mentions = list()
        for i in range(len(source)):
            mentions.append(source[i].text + " </s> " + target[i].text)
        embeddings = self.model.model_body.encode(mentions, normalize_embeddings=self.model.normalize_embeddings, convert_to_tensor=True)
        predicts = self.model.model_head.predict_proba(embeddings)
        for j in range(len(predicts[0])):
            predscores = {self.labels[i]: arr[:,1].tolist()[j] for i, arr in enumerate(predicts)}
            top_n_labels = sorted(predscores, key=predscores.get, reverse=True)[:self.n]
            filtered_labels = [label for label in top_n_labels if predscores[label] > self.threshold]
            final_labels.append(filtered_labels)
        return final_labels


from sklearn.cluster import KMeans
import numpy as np
from transformers import AutoTokenizer, AutoModel
import torch
import pickle
from .Categorizer import Categorizer

class Clustering(Categorizer):
    def __init__(self, n_clusters, model_path, output_path, clustering_model):
        """
        Initializes the Clustering class.

        Parameters:
        n_clusters (int): Number of clusters to create.
        model_path (str): Path to the pretrained clustering model (can be None).
        output_path (str): Path to save the trained clustering model.
        clustering_model (str): Name of the clustering model.
        """
        self.tokenizer = AutoTokenizer.from_pretrained(model_path)
        self.model = AutoModel.from_pretrained(model_path)
        self.trained = False
        self.n_clusters = n_clusters
        self.output_path = output_path
        self.clustering_model = clustering_model
        
    def generate_embeddings(self, mentions):
        """
        Generates embeddings for a list of text mentions.

        Parameters:
        mentions (list): List of text mentions to generate embeddings for.

        Returns:
        torch.Tensor: Tensor containing the generated embeddings.
        """
        tokenized_inputs = self.tokenizer(mentions, padding=True, truncation=True, return_tensors="pt")
        with torch.no_grad():
            outputs = self.model(**tokenized_inputs)
        embeddings = outputs.last_hidden_state
        return embeddings[:, 0, :]

    def train_clusters(self, mentions_train, **kwargs):
        """
        Trains a clustering model using the provided text mentions.

        Parameters:
        mentions_train (list): List of text mentions for training.
        **kwargs: Additional arguments for KMeans clustering.

        Prints:
        - A warning message for specified parameters that cannot be changed when that is attempted.
        - Examples of mentions in each cluster (no more than 5 each).
        """
        mentions_embeddings_train = self.generate_embeddings(mentions_train)
        specified_params = ['random_state', 'n_init']
        for param in specified_params:
            if param in kwargs:
                del kwargs[param]
                print("Warning. The parameter " + param + " cannot be changed from its given value.")
        self.kmeans = KMeans(n_clusters=self.n_clusters, random_state=0, n_init="auto", **kwargs).fit(mentions_embeddings_train)
        self.trained = True
        self.save_trained_model()

        clusters_examples = dict()
        for i in range(max(self.kmeans.labels_)+1):
            instances = []
            for j in range(len(self.kmeans.labels_)):
                if self.kmeans.labels_[j] == i and len(instances) < 5:
                    instances.append(mentions_train[j])
            clusters_examples["Class " + str(i)] = instances
        print(clusters_examples)

    def predict_clusters(self, mentions):
        """
        Predicts clusters for a list of text mentions.

        Parameters:
        mentions (list): List of text mentions for cluster prediction.

        Returns:
        dict: Dictionary mapping mentions to their predicted clusters.
        """
        word_with_cluster = {}
        mentions_embeddings = self.generate_embeddings(mentions)
        if self.clustering_model is not None and not self.trained:
            self.import_pretrained_model(self.clustering_model)
            self.trained = True
        elif not self.trained:
            self.train_clusters(mentions)
            self.trained = True
            self.save_trained_model()
        predicted_clusters = self.kmeans.predict(mentions_embeddings)
        for i in range(len(mentions)):
            word_with_cluster[mentions[i]] = predicted_clusters[i]
        return word_with_cluster
    
    def save_trained_model(self):
        """
        Saves the trained KMeans clustering model to the specified output path.
        """
        with open(self.output_path, 'wb') as file:
            pickle.dump(self.kmeans, file)      

    def import_pretrained_model(self, clustering_model):
        """
        Imports a pretrained clustering model from a file.

        Parameters:
        model_path (str): Path to the pretrained clustering model file.
        """
        with open(clustering_model, 'rb') as file:
            self.kmeans = pickle.load(file)
        self.trained = True



import os
import re
import nltk
nltk.download('stopwords')

class Extractor:
    def __init__(self, language, max_tokens, top_n):
        """
        Initializes an Extractor object with specified parameters.

        Parameters:
        language (str): Language for text processing.
        max_tokens (int): Maximum number of tokens for extracted terms.
        top_n (int): number of keywords to be extracted.

        Returns:
        None
        """
        self.language = language
        self.max_tokens = max_tokens
        self.top_n = top_n

    def extract_terms(self, text):
        """
        This method must be implemented in a subclass to extract terms from the given text.

        Parameters:
        text (str): Input text for term extraction.

        Raises:
        NotImplementedError: When called directly from the base class.

        Returns:
        None
        """
        raise NotImplementedError("extract_terms method must be implemented in subclass")

    def extract_terms_with_span(self, text):
        """
        Extracts terms from the given text and associates them with their spans.

        Parameters:
        text (str): Input text for term extraction.

        Returns:
        list: List of extracted terms with spans.
        """
        terms = self.extract_terms(text)
        terms_with_span = self.find_term_span(text, terms)
        return terms_with_span

    def extract_terms_without_overlaps(self, text, kwargs):
        """
        Extracts terms from the given text with their span and removing complete overlaps.

        Parameters:
        text (str): Input text for term extraction.
        **kwargs: Additional arguments for training. If some of the kwargs for the extraction do not exist, they will be ignored.

        Returns:
        list: List of extracted terms without overlaps.
        """
        self.kwargs = kwargs
        terms_with_span = self.extract_terms_with_span(text)
        terms_without_overlaps = self.rmv_overlaps(terms_with_span)
        return terms_without_overlaps

    def postprocess_terms(self, terms):
        """
        Post-processes the extracted terms if needed by removing and modifying some terms in a specific order.

        Parameters:
        terms (list): List of extracted terms.

        Returns:
        new_terms (list): List of extracted terms after postprocessing.
        """
        new_terms = self.rmv_meaningless(terms)
        new_terms = self.rmv_stopwords(new_terms)
        new_terms = self.rmv_stopwords(new_terms)
        new_terms = self.rmv_nonalnum_characters(new_terms)
        new_terms = self.rmv_nonalnum_characters(new_terms)
        new_terms = self.rmv_meaningless(new_terms)
        return new_terms

    def rmv_meaningless(self, terms):
        """
        Removes meaningless terms from the list of terms (numbers, terms with just one character and stopwords).

        Parameters:
        terms (list): List of terms to be filtered.

        Returns:
        list: List of terms with meaningless terms removed.
        """
        self.stopwords = nltk.corpus.stopwords.words(self.language)
        new_terms = [t for t in terms if (t[0] not in self.stopwords and len(t[0]) >= 2 and not t[0].isdigit())]
        return new_terms

    def rmv_stopwords(self, terms):
        """
        Removes stopwords at the start and end of terms from the list.

        Parameters:
        terms (list): List of terms to be filtered.

        Returns:
        list: List of terms with stopwords removed.
        """
        new_terms = terms
        self.stopwords = nltk.corpus.stopwords.words(self.language)
        for word in self.stopwords:
            new_terms = [(t[0][(len(word)+1):], t[1] + len(word) + 1, t[2], t[3], t[4]) if (t[0].lower().startswith(word.lower() + " ")) else t for t in new_terms]
            new_terms = [(t[0][:-(len(word)+1)], t[1], t[2] - len(word) - 1, t[3], t[4]) if (t[0].lower().endswith(" " + word.lower())) else t for t in new_terms]
        return new_terms

    def rmv_nonalnum_characters(self, terms):
        """
        Removes non-alphanumeric characters at the start and end of the terms.

        Parameters:
        terms (list): List of terms to be filtered.

        Returns:
        list: List of terms with non-alphanumeric characters removed.
        """
        new_terms = []
        for t in terms:
            if (any(char.isalnum() for char in t[0])):
                new_terms.append(t)
        new_terms = [(t[0][1:], t[1] + 1, t[2], t[3], t[4]) if not t[0][0].isalnum() else t for t in new_terms]
        new_terms = [(t[0][:-1], t[1], t[2] - 1, t[3], t[4]) if not t[0][-1].isalnum() else t for t in new_terms]
        return new_terms

    def select_top_n(self, text):
        """
        Computes the number of keywords to be extracted from a document when 'auto' is selected or raises an error if the given value is not an integer.

        Parameters:
        text (str): Text from which the keywords are extracted.
        """
        if self.top_n is None:
            words = [w for w in text.split() if w != ""]
            self.top_n = round(len(words)/10)
        elif not isinstance(self.top_n, int):
            raise TypeError("top_n variable must be type int.")

    @staticmethod
    def find_term_span(text, terms):
        """
        Finds the span of each term in the text.

        Parameters:
        text (str): Input text for term extraction.
        terms (list): List of extracted terms.

        Returns:
        list: List of term spans including term, start, end and score.
        """
        spans = []
        for t, score in terms:
            term = re.escape(t)
            patron = r'\b' + term + r'\b'
            coincidencias = re.finditer(patron, text, re.IGNORECASE)
            span = [(t, coincidencia.start(), coincidencia.end(), score) for coincidencia in coincidencias]
            spans.extend(span)
        return spans

    @staticmethod
    def rmv_overlaps(keywords):
        """
        Removes completely overlapping terms from the list of extracted keywords.

        Parameters:
        keywords (list): List of extracted keywords with spans and other attributes (score, method, etc).

        Returns:
        list: List of non-overlapping keywords.
        """
        ent = [kw[0] for kw in keywords]
        pos = [kw[1:3] for kw in keywords]
        score = [kw[3:] for kw in keywords]
        updated_keywords = []
        repeated_words = []
        for i in range(len(ent)):
            overlap = False
            if pos[i] in repeated_words:
                overlap = True
            else:
                repeated_words.append(pos[i])
                for k in range(len(ent)):
                    if (((pos[i][0] >= pos[k][0]) and (pos[i][1] < pos[k][1])) or ((pos[i][0] > pos[k][0]) and (pos[i][1] <= pos[k][1]))):
                        overlap = True
            if (not overlap):
                kw = (ent[i], pos[i][0], pos[i][1]) + score[i]
                updated_keywords.append(kw)
        return updated_keywords


from keybert import KeyBERT
from sentence_transformers import SentenceTransformer
import os
import nltk
nltk.download('stopwords')
from keyphrase_vectorizers import KeyphraseCountVectorizer
from .Extractor import Extractor

class KeyBertExtractor(Extractor):
    def __init__(self, language, max_tokens, pos, pos_pattern, top_n):
        """
        Initializes the KeyBertExtractor object with specified parameters and the KeyBERT object itself.

        Parameters:
        language (str): Language for keyword extraction.
        max_tokens (int): Maximum number of tokens in extracted keywords.
        pos (bool): Wether to use Part of Speech sequences.
        pos_pattern (str): The Part of Speech regex pattern.
        top_n (int): number of keywords to be extracted.
        """
        super().__init__(language, max_tokens, top_n)
        sentence_model = self.initialize_sentence_model()
        self.extractor = KeyBERT(model=sentence_model)
        self.pos = pos
        self.pos_pattern = pos_pattern
        self.top_n = top_n

    def initialize_sentence_model(self):
        """
        Initializes the sentence embedding model.

        Parameters:
        None.

        Returns:
        SentenceTransformer: Initialized sentence embedding model.
        """
        path = 'BSC-NLP4BIA/SapBERT-parents-from-roberta-base-biomedical-clinical-es'
        sentence_model = SentenceTransformer(path)
        return sentence_model

    def extract_terms(self, text):
        """
        Extracts keywords from the given text using KeyBERT (with PoS or with n_gram_range).

        Parameters:
        text (str): Input text for keyword extraction.

        Returns:
        list: List of extracted keywords and their scores, sorted by score in descending order.
        """
        self.select_top_n(text)
        filtered_kwargs = {key: value for key, value in self.kwargs.items() if key in self.extractor.extract_keywords.__code__.co_varnames}
        specified_params = ['keyphrase_ngram_range', 'stop_words', 'use_mmr']
        for param in specified_params:
            if param in filtered_kwargs:
                del filtered_kwargs[param]
                print("Warning. The parameter " + param + " can not be changed from its given value.")
        keywords_all = list()
        if self.pos:
            keywords_all = self.extractor.extract_keywords(text, vectorizer=KeyphraseCountVectorizer(spacy_pipeline="es_core_news_sm", stop_words='spanish', pos_pattern=self.pos_pattern), top_n=self.top_n, **filtered_kwargs)
        else:
            for i in range(self.max_tokens):
                keywords_i = self.extractor.extract_keywords(text, keyphrase_ngram_range=(1, i+1), stop_words=None, use_mmr=True, top_n=self.top_n, **filtered_kwargs)
                keywords_all.extend(keywords_i)
        terms = sorted(keywords_all, key=lambda x: x[1], reverse=True)
        return terms



import os
import re
from nltk.tokenize import word_tokenize
from .extractors.RakeExtractor import RakeExtractor
from .extractors.YakeExtractor import YakeExtractor
from .extractors.TextRankExtractor import TextRankExtractor
from .extractors.KeyBertExtractor import KeyBertExtractor
from .utils.data_structures import Keyword
from .categorizers.Clustering import Clustering
from .categorizers.SetFitClassifier import SetFitClassifier
from .categorizers.TransformersClassifier import TransformersClassifier

class TermExtractor:
    def __init__(self,
                 extraction_methods=["textrank"], 
                 categorization_method="setfit", 
                 language="spanish", 
                 max_tokens=3, 
                 pos=False,
                 pos_pattern="<NOUN.*>*<ADP.*>*<NOUN.*>*<ADJ.*>*|<PROPN.*>+|<VERB.*>+",
                 top_n=None,
                 join=False, 
                 postprocess=True,
                 n=1, 
                 thr_setfit=0.5,
                 thr_transformers=-1,
                 n_clusters=None,
                 categorizer_model_path="BSC-NLP4BIA/SapBERT-from-roberta-base-biomedical-clinical-es",
                 output_path="./trained_model", 
                 clustering_model=None,
                 classifier_model=None,
                 **kwargs,
                ):
        """
        Initializes the TextCategorization Pipeline.

        Parameters:
        extraction_methods (list): List of keyword extraction methods.
        categorization_method (str): Method for text categorization.
        language (str): Language for text processing.
        max_tokens (int): Maximum number of tokens for keyword extraction.
        pos (bool): Whether to use Part of Speech sequences in KeyBert.
        pos_pattern (str): The Part of Speech regex pattern.
        top_n (int): The number of keywords to extract from each document.
        join (bool): Whether to join keywords from different methods and remove overlaps among them.
        postprocess (bool): Whether to apply postprocessing to keywords.
        n (int): Maximum number of labels for a single sample.
        thr_setfit (float): Threshold for SetFit classification.
        thr_transformers (float): Threshold for Transformers classification.
        n_clusters (int): Number of clusters for clustering.
        categorizer_model_path (str): Path to the base categorizer model.
        output_path (str): Path to save the trained model.
        clustering_model (str): Clustering model for keyword clustering.
        classifier_model (str): Classifier model for classification.
        **kwargs: Additional keyword arguments.
        """
        self.extraction_methods = extraction_methods
        self.extractors = self.initialize_keyword_extractors(language, max_tokens, pos, pos_pattern, top_n)
        self.categorization_method = categorization_method
        self.categorizer = self.initialize_categorizers(n, thr_setfit, thr_transformers, n_clusters, categorizer_model_path, output_path, clustering_model, classifier_model)
        self.join = join
        self.postprocess = postprocess
        self.kwargs = kwargs
        self.keywords = None
    
    def __call__(self, text):
        """
        Extracts terms from the given text using initialized extractors and categorize them using initialized categorizers.

        Parameters:
        text (str): Input text for term extraction.

        Returns:
        None
        """
        self.extract_terms(text, self.join, self.postprocess)
        self.categorize_terms()

    def initialize_keyword_extractors(self, language, max_tokens, pos, pos_pattern, top_n):
        """
        Initializes keyword extractors based on selected extraction methods.

        Parameters:
        language (str): Language for text processing.
        max_tokens (int): Maximum number of tokens for extracted terms.
        pos (bool): Wether to use Part of Speech sequences in KeyBert.
        pos_pattern (str): The Part of Speech regex pattern.
        top_n (int): number of keywords to be extracted.

        Returns:
        dict: Dictionary containing initialized keyword extractors.
        """
        keyword_extractors = {}
        
        if 'rake' in self.extraction_methods:
            keyword_extractors["rake"] = RakeExtractor(language, max_tokens, top_n)
        
        if 'yake' in self.extraction_methods:
            keyword_extractors["yake"] = YakeExtractor(language, max_tokens, top_n)
        
        if 'textrank' in self.extraction_methods:
            keyword_extractors["textrank"] = TextRankExtractor(language, max_tokens, top_n)

        if 'keybert' in self.extraction_methods:
            keyword_extractors["keybert"] = KeyBertExtractor(language, max_tokens, pos, pos_pattern, top_n)

        if not keyword_extractors:
            raise ValueError("No extraction method called {}".format(self.extraction_methods))
        
        return keyword_extractors

    def extract_terms(self, text, join, postprocess):
        """
        Extracts terms from the given text using initialized extractors.

        Parameters:
        text (str): Input text for term extraction.
        join (bool): Whether to join terms obtained with different methods and to remove overlaps among them.
        postprocess (bool): Whether to postprocess the extracted keywords to enhance them.

        Returns:
        list: List of extracted terms.
        """
        try:
            all_terms = []
            for key, extractor in self.extractors.items():
                terms = extractor.extract_terms_without_overlaps(text, kwargs=self.kwargs)
                terms = [term + (key,) for term in terms]
                all_terms.extend(terms)
            if join:
                all_terms = list(self.extractors.values())[0].rmv_overlaps(all_terms)
            if postprocess:
                all_terms = list(self.extractors.values())[0].postprocess_terms(all_terms)
                all_terms = list(self.extractors.values())[0].rmv_overlaps(all_terms)
            self.keywords = [Keyword(text=i[0], extraction_method=i[4], ini=i[1], fin=i[2], score=i[3]) for i in all_terms]
        except:
            raise AttributeError("A list of extractors must be provided")

    def initialize_categorizers(self, n, thr_setfit, thr_transformers, n_clusters, model_path, output_path, clustering_model, classifier_model):
        """
        Initializes the categorizer based on the selected method.

        Parameters:
        n (int): Maximum number of labels for a single mention.
        thr_setfit (float): Threshold for label filtering in SetFitClassifier.
        thr_transformers (float): Threshold for label filtering in TransformersClassifier.
        n_clusters (int): Number of clusters for clustering-based categorization.
        model_path (str): Path to the pretrained model.
        output_path (str): Path to save the trained model.
        clustering_model (str): Clustering model for keyword clustering.
        classifier_model (str): Classifier model for keyword classification.

        Returns:
        Categorizer: Initialized categorizer object.
        """
        if 'transformers' == self.categorization_method:
            categorizer = TransformersClassifier(n, thr_transformers, model_path, output_path, classifier_model)
            
        elif 'setfit' == self.categorization_method:
            categorizer = SetFitClassifier(n, thr_setfit, model_path, output_path, classifier_model)

        elif 'clustering' == self.categorization_method:
            if n_clusters is None:
                raise TypeError("TermExtractor.__init__() missing 1 required positional argument: 'n_clusters' when selecting the Clustering algorithm")
            categorizer = Clustering(n_clusters, model_path, output_path, clustering_model)
        else:
            raise ValueError("No categorizer method called {}".format(self.categorization_method))
        return categorizer

    def categorize_terms(self):
        """
        Categorizes the extracted terms (keyword objects) using the selected categorizer method.
        """
        try:
            if self.categorization_method == 'clustering':
                list_of_keywords = [kw.text for kw in self.keywords]
                clusters = self.categorizer.predict_clusters(list_of_keywords)
                for kw in self.keywords:
                    kw.label = clusters[kw.text]
                    kw.categorization_method = self.categorization_method
            else:
                labels = self.categorizer.compute_predictions(self.keywords)
                for i in range(len(self.keywords)):
                    self.keywords[i].label = labels[i]
                    self.keywords[i].categorization_method = self.categorization_method
        except:
            raise AttributeError("A categorizer method must be provided")

    def train_classifier(self, trainX, trainY, testX, testY, mcm=False, classification_report=False, **kwargs):
        """
        Trains and evaluates the text categorization model.

        Parameters:
        trainX (list): List of training data (mentions).
        trainY (list): List of training labels.
        testX (list): List of test data (mentions).
        testY (list): List of test labels.
        mcm (bool): Whether to create a multi-label confusion matrix heatmap.
        classification_report (bool): Whether to display a classification report.
        **kwargs: Additional keyword arguments.

        Prints:
        - String: Metrics to evaluate the performance of the catgorizer.
        """
        self.categorizer.initialize_model_body(trainY)
        metrics = self.categorizer.train_evaluate(trainX, trainY, testX, testY, mcm, classification_report, **kwargs)
        print(metrics)

    def train_clustering(self, trainX, **kwargs):
        """
        Trains the text clustering model using the provided data.

        Parameters:
        trainX (list): List of training data (mentions).
        **kwargs: Additional keyword arguments.
        """
        self.categorizer.train_clusters(trainX, **kwargs)


class Keyword:
    def __init__(self, text, extraction_method=None, ini=None, fin=None, score=None, label=None, categorization_method=None):
        """
        Initializes a Keyword object with the provided attributes.

        Parameters:
        text (str): The extracted keyword text. It must be provided.
        extraction_method (str): The extraction method used to find the keyword. Default: None.
        ini (int): The starting index of the keyword in the input text. Default: None.
        fin (int): The ending index of the keyword in the input text. Default: None.
        score (float): The relevance score of the keyword. Default: None.
        label (list): List of labels associated with the mention. Default: None.
        categorization_method (str): The categorization menthod used to categorize the keyword. Default: None.
        """
        self.text = text
        self.extraction_method = extraction_method
        self.score = score
        self.span = [ini, fin]
        self.label = label
        self.categorization_method = categorization_method

    def __repr__(self):
        """
        Returns a string representation of the Keyword object.

        Returns:
        str: A string representing the Keyword object.
        """
        return f"<Keyword(text='{self.text}', span='{self.span}', extraction method='{self.extraction_method}', score='{self.score}', categorization method='{self.categorization_method}', class='{self.label}')>"


class Relation:
    def __init__(self, source, target, rel_type, relation_method):
        """
        Initializes the Relation class.

        Parameters:
        source (Keyword): Source mention for the relation.
        target (Keyword): Target mention for the relation.
        rel_type (str): Type of relation.
        relation_method (str): Method used to compute the relation.
        """
        self.source = source
        self.target = target
        self.rel_type = rel_type
        self.relation_method = relation_method

    def __repr__(self):
        """
        String representation of Relation object.

        Returns:
        str: Representation of the Relation object.
        """
        return f"<Relation(source mention='{self.source.text}', target mention='{self.target.text}', relation type='{self.rel_type}', relation method='{self.relation_method}')>"



import nltk
nltk.download('stopwords')
nltk.download('punkt')
from nltk.tokenize import word_tokenize
from rake_nltk import Rake
from .Extractor import Extractor

class RakeExtractor(Extractor):
    def __init__(self, language, max_tokens, top_n):
        """
        Initializes a RakeExtractor object with specified parameters and the Rake object itself.

        Parameters:
        language (str): Language for text processing.
        max_tokens (int): Maximum number of tokens for extracted terms.
        top_n (int): number of keywords to be extracted.

        Returns:
        None
        """
        super().__init__(language, max_tokens, top_n)
        self.stopwords = nltk.corpus.stopwords.words(language)
        self.extractor = Rake(stopwords=self.stopwords, language=language)

    def extract_terms(self, text):
        """
        Extracts terms using the RAKE algorithm and the specified parameters.

        Parameters:
        text (str): Input text for term extraction.

        Returns:
        list: List of extracted terms with scores.
        """
        self.select_top_n(text)
        filtered_kwargs = {key: value for key, value in self.kwargs.items() if key in self.extractor.extract_keywords_from_text.__code__.co_varnames}
        self.extractor.extract_keywords_from_text(text, **filtered_kwargs)
        terms = self.extractor.get_ranked_phrases_with_scores()
        terms = [(kw,score) for score,kw in terms if (len(word_tokenize(kw, language=self.language)) <= self.max_tokens)]
        sorted_terms = sorted(terms, key=lambda x: x[1], reverse=True)
        try:
            return sorted_terms[:self.top_n]
        except IndexError:
            return sorted_terms


import yake
from nltk.tokenize import word_tokenize
from .Extractor import Extractor

class YakeExtractor(Extractor):
    def __init__(self, language, max_tokens, top_n):
        """
        Initializes a YakeExtractor object with specified parameters.

        Parameters:
        language (str): Language for text processing.
        max_tokens (int): Maximum number of tokens for extracted terms.
        top_n (int): number of keywords to be extracted.
        """
        super().__init__(language, max_tokens, top_n)
        if language == 'spanish':
            self.extractor = yake.KeywordExtractor() 
        else:
            raise ValueError("Expected spanish language. Other languages not recognized")

    def extract_terms(self, text):
        """
        Extracts terms using the YAKE algorithm.

        Parameters:
        text (str): Input text for term extraction.

        Returns:
        list: List of extracted terms with scores.
        """
        self.select_top_n(text)
        filtered_kwargs = {key: value for key, value in self.kwargs.items() if key in self.extractor.extract_keywords.__code__.co_varnames}
        keywords = self.extractor.extract_keywords(text, **filtered_kwargs)
        terms = [(kw,score) for kw, score in keywords if (len(word_tokenize(kw)) <= self.max_tokens)]
        sorted_terms = sorted(terms, key=lambda x: x[1], reverse=True)
        try:
            return sorted_terms[:self.top_n]
        except IndexError:
            return sorted_terms


class Relator:
    def __init__(self, n, threshold, model_path):
        """
        Initializes the Relator class.

        Parameters:
        n (int): Maximum number of labels for a single relation.
        threshold (float): Threshold value used in the relation extraction.
        model_path (str): Path to the model.
        """
        self.n = n
        self.threshold = threshold
        self.labels = ['BROAD','EXACT','NARROW','NO_RELATION']
        self.model = self.initialize_pretrained_model(model_path)


from setfit import SetFitModel, SetFitTrainer
import torch
import pandas as pd
from datasets import Dataset
from .Categorizer import Categorizer
from sklearn.preprocessing import MultiLabelBinarizer

class SetFitClassifier(Categorizer):
    def __init__(self, n, threshold, model_path, output_path, classifier_model):
        """
        Initializes the SetFitClassifier class.

        Parameters:
        n (int): Maximum number of labels for a single mention.
        threshold (float): Threshold for label prediction.
        model_path (str): Path to the pretrained model (can be None).
        output_path (str): Path to save the trained model.
        classifier_model (str): Name of the classifier model.
        """
        super().__init__(n, threshold, model_path, output_path, classifier_model)

    def initialize_pretrained_model(self):
        """
        Initializes and returns a pretrained and finetuned model (default or specified).
        
        Returns:
        SetFitModel: The pretrained model.
        """
        if self.classifier_model is None:
            path = 'BSC-NLP4BIA/biomedical-term-classifier-setfit'
            model = SetFitModel.from_pretrained(path)
        else:
            model = SetFitModel.from_pretrained(self.classifier_model)
        return model

    def compute_predictions(self, mention):
        """
        Computes label predictions for a list of given mentions.

        Parameters:
        mention (list): List of input mentions for prediction.

        Returns:
        list: List of lists of filtered labels considering the given threshold and the maximum labels.
        """
        final_labels = list()
        mention_text = [m.text for m in mention]
        embeddings = self.model.model_body.encode(mention_text, normalize_embeddings=self.model.normalize_embeddings, convert_to_tensor=True)
        predicts = self.model.model_head.predict_proba(embeddings)
        for j in range(len(predicts[0])):
            predscores = {self.labels[i]: arr[:,1].tolist()[j] for i, arr in enumerate(predicts)}
            top_n_labels = sorted(predscores, key=predscores.get, reverse=True)[:self.n]
            filtered_labels = [label for label in top_n_labels if predscores[label] > self.threshold]
            final_labels.append(filtered_labels)
        return final_labels

    def initialize_model_body(self, trainY):
        """
        Initializes the model body from the given base model.

        Parameters: trainY
        """
        self.model = SetFitModel.from_pretrained(self.model_path, multi_target_strategy="multi-output")

    def train_evaluate(self, trainX, trainY, testX, testY, mcm, classification_report, **kwargs):
        """
        Trains and evaluates the SetFit model on the provided data.

        Parameters:
        trainX (list): List of training data (mentions).
        trainY (list): List of training labels.
        testX (list): List of test data (mentions).
        testY (list): List of test labels.
        mcm (bool): Whether to create a multi-label confusion matrix heatmap.
        classification_report (bool): Whether to display a classification report.
        **kwargs: Additional arguments for training.

        Prints:
        - A warning message for specified parameters that cannot be changed when that is attempted.
        """
        train_dataset, test_dataset = self.prepare_data(trainX, trainY, testX, testY)
        evaluate_with_params = self.lambda_evaluate_model(mcm, classification_report)
        specified_params = ['metric', 'num_iterations']
        for param in specified_params:
            if param in kwargs:
                del kwargs[param]
                print("Warning. The parameter " + param + " cannot be changed from its given value.")
        trainer = SetFitTrainer(model=self.model, 
            train_dataset=train_dataset, 
            eval_dataset=test_dataset, 
            metric=evaluate_with_params, 
            num_iterations=5, 
            **kwargs,
        )
        trainer.train()
        metrics = trainer.evaluate()
        self.model.save_pretrained(self.output_path)
        return metrics

    def prepare_data(self, trainX, trainY, testX, testY):
        """
        Prepares training and test datasets for SetFit model training.

        Parameters:
        trainX (list): List of training data (mentions).
        trainY (list): List of training labels.
        testX (list): List of test data (mentions).
        testY (list): List of test labels.

        Returns:
        Tuple: Tuple containing training and test datasets.
        """
        trainY = [{i} for i in trainY]
        testY = [{i} for i in testY]
        mlb = MultiLabelBinarizer()
        mlb.fit_transform(trainY)
        self.labels = [i for i in mlb.classes_]
        train_dataset = Dataset.from_dict({"text": trainX, "label": mlb.fit_transform(trainY)})
        test_dataset = Dataset.from_dict({"text": testX, "label": mlb.transform(testY)})
        return train_dataset, test_dataset



from itertools import product
from .utils.data_structures import Relation
from .utils.data_structures import Keyword
from .relators.SetFitRelator import SetFitRelator
from .relators.TransformersRelator import TransformersRelator

class RelExtractor:
    def __init__(self, 
                 relation_method='transformers', 
                 language='spanish',
                 n=1,
                 thr_transformers=-1,
                 thr_setfit=0.5,
                 all_combinations=False,
                 model_path=None,
                ):
        """
        Initializes the RelExtractor class.

        Parameters:
        relation_method (str): Method used for relation extraction.
        language (str): Language for text processing.
        n (int): Maximum number of labels for a single relation.
        thr_setfit (float): Threshold for SetFit relation classification.
        thr_transformers (float): Threshold for Transformers relation classification.
        all_combinations (bool): Whether to create relations for all combinations of source and target.
        model_path (str): Path to the model if required.

        """
        self.relation_method=relation_method
        self.all_combinations = all_combinations
        self.rel_extractor = self.initialize_relation_method(language, n, thr_transformers, thr_setfit, model_path)
        
    def initialize_relation_method(self, language, n, thr_transformers, thr_setfit, model_path):
        """
        Initializes the selected relation extraction method.

        Parameters:
        language (str): Language for text processing.
        n (int): Maximum number of labels for a single relation.
        thr_setfit (float): Threshold for SetFit relation classification.
        thr_transformers (float): Threshold for Transformers relation classification.
        model_path (str): Path to the model if required.

        Returns:
        object: Instance of the selected relation extraction method.
        """
        if 'transformers' == self.relation_method:
            rel_extractor = TransformersRelator(n, thr_transformers, model_path)
        elif 'setfit' == self.relation_method:
            rel_extractor = SetFitRelator(n, thr_setfit, model_path)
        else:
            raise ValueError("No relation method called {}".format(self.relation_method))
        return rel_extractor
    
    def __call__(self, source, target):
        """
        Executes the relation extraction based on the source and target inputs.

        Parameters:
        source (str/Keyword/list): Source for relation extraction.
        target (str/Keyword/list): Target for relation extraction.

        Raises:
        TypeError: If source or target are not of expected types or lengths.

        Returns:
        list: List of Relation objects based on the extraction.
        """
        if (type(source)==Keyword):
            source = [source]
        elif (type(source)==str):
            source = [Keyword(text=source)]
        elif (type(source)==list):
            if (all(isinstance(element, str) for element in source)):
                source = [Keyword(text=i) for i in source]
            elif not (all(isinstance(element, Keyword) for element in source)):
                raise TypeError('Source contains elements other than strings or Keyword class objects')
        else:
            raise TypeError('Source must be a string, a Keyword class object, a list of strings or a list of Keyword class objects')

        if (type(target)==Keyword):
            target = [target]
        elif (type(target)==str):
            target = [Keyword(text=target)]
        elif (type(target)==list):
            if (all(isinstance(element, str) for element in target)):
                target = [Keyword(text=i) for i in target]
            elif not (all(isinstance(element, Keyword) for element in target)):
                raise TypeError('Target contains elements other than strings or Keyword class objects')
        else:
            raise TypeError('Target must be a string, a Keyword class object, a list of strings or a list of Keyword class objects')

        if self.all_combinations:
            source_rep = [i for i,j in list(product(source, target))]
            target_rep = [j for i,j in list(product(source, target))]
            relations = self.rel_extractor.compute_relation(source_rep, target_rep)
            self.relations = [Relation(source_rep[i],target_rep[i],relations[i], self.relation_method) for i in range(len(source_rep))]
        else:
            if (len(source) == len(target)):
                relations = self.rel_extractor.compute_relation(source, target)
                self.relations = [Relation(source[i],target[i],relations[i], self.relation_method) for i in range(len(source))]
            else:
                raise TypeError('Source and target must be the same length when all_combinations=False.')


import torch
from sklearn.preprocessing import MultiLabelBinarizer
from transformers import AutoTokenizer, AutoModelForSequenceClassification
from .Relator import Relator

class TransformersRelator(Relator):
    def __init__(self, n, threshold, model_path):
        """
        Initializes TransformersRelator, a class inheriting from Relator.

        Parameters:
        n (int): Maximum number of labels for a single relation.
        threshold (float): Threshold value used for mentions relation.
        model_path (str): Path to the model class.
        """
        super().__init__(n, threshold, model_path)
        
    def initialize_pretrained_model(self, model_path):
        """
        Initializes a pretrained model for TransformersRelator based on the provided model_path.

        Parameters:
        model_path (str): Path to the pretrained model.

        Returns:
        object: Pretrained model instance.
        """
        self.mlb = MultiLabelBinarizer()
        self.mlb.fit([self.labels])
        path = 'BSC-NLP4BIA/SapBERT-parents-from-roberta-base-biomedical-clinical-es'
        self.tokenizer = AutoTokenizer.from_pretrained(path)
        if model_path is None:
            path = 'BSC-NLP4BIA/biomedical-semantic-relation-classifier'        
            model = AutoModelForSequenceClassification.from_pretrained(path)
        else:
            model = AutoModelForSequenceClassification.from_pretrained(model_path)
        return model

    def compute_relation(self, source, target):
        """
        Computes relations between source and target entities using Transformers model.

        Parameters:
        source (list): List of source entities.
        target (list): List of target entities.

        Returns:
        list: List of labels representing computed relations.
        """
        final_labels = list()
        source_text = [s.text for s in source]
        target_text = [t.text for t in target]
        tokenized_mention = self.tokenizer(source_text, target_text, return_tensors='pt', padding=True, truncation=True)
        with torch.no_grad():
            output = self.model(**tokenized_mention)
        logits = output.logits
        for i in range(len(logits.tolist())):
            predscores = {label: score for label, score in zip(self.labels, logits.tolist()[i])}
            top_n_labels = sorted(predscores, key=predscores.get, reverse=True)[:self.n]
            filtered_labels = [label for label in top_n_labels if predscores[label] > self.threshold]
            final_labels.append(filtered_labels)
        return final_labels


import pandas as pd
import torch
import numpy as np
from torch.utils.data import TensorDataset
from sklearn.preprocessing import MultiLabelBinarizer
from transformers import AutoTokenizer, AutoModelForSequenceClassification, Trainer, TrainingArguments
from .Categorizer import Categorizer

class TransformersClassifier(Categorizer):
    def __init__(self, n, threshold, model_path, output_path, classifier_model):
        """
        Initializes the TransformersClassifier class.

        Parameters:
        n (int): Maximum number of labels for a single mention.
        threshold (float): Threshold for label prediction.
        model_path (str): Path to the pretrained model (can be None).
        output_path (str): Path to save the trained model.
        classifier_model (str): Name of the classifier model.
        """
        super().__init__(n, threshold, model_path, output_path, classifier_model)
        
    def initialize_pretrained_model(self):
        """
        Initializes and returns a pretrained and fine-tuned model.

        Returns:
        AutoModelForSequenceClassification: The pretrained model.
        """
        self.mlb = MultiLabelBinarizer()
        self.mlb.fit([self.labels])
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_path)
        if self.classifier_model is None:
            path = 'BSC-NLP4BIA/biomedical-term-classifier'        
            model = AutoModelForSequenceClassification.from_pretrained(path)
        else:
            model = AutoModelForSequenceClassification.from_pretrained(self.classifier_model)
        return model

    def compute_predictions(self, mention):
        """
        Computes label predictions for a list of given mentions.

        Parameters:
        mention (list): List of input mentions for prediction.

        Returns:
        list: List of lists of filtered labels considering the given threshold and the maximum labels.
        """
        final_labels = list()
        mention_text = [m.text for m in mention]
        tokenized_mention = self.tokenizer(mention_text, return_tensors='pt', padding=True, truncation=True)
        with torch.no_grad():
            output = self.model(**tokenized_mention)
        logits = output.logits
        for i in range(len(logits.tolist())):
            predscores = {label: score for label, score in zip(self.labels, logits.tolist()[i])}
            top_n_labels = sorted(predscores, key=predscores.get, reverse=True)[:self.n]
            filtered_labels = [label for label in top_n_labels if predscores[label] > self.threshold]
            final_labels.append(filtered_labels)
        return final_labels
    
    def initialize_model_body(self, trainY):
        """
        Initializes the model body for training. That includes finding labels, initializing the tokenizer, and the model for multi-label classification

        Parameters:
        trainY (list): List of training labels.
        """
        self.find_labels(trainY)
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_path)
        self.model = AutoModelForSequenceClassification.from_pretrained(self.model_path, num_labels=self.num_labels, problem_type="multi_label_classification")

    def train_evaluate(self, trainX, trainY, testX, testY, mcm, classification_report, **kwargs):
        """
        Trains and evaluates the Transformers model on the provided data.

        Parameters:
        trainX (list): List of training data (mentions).
        trainY (list): List of training labels.
        testX (list): List of test data (mentions).
        testY (list): List of test labels.
        mcm (bool): Whether to create a multi-label confusion matrix heatmap.
        classification_report (bool): Whether to display a classification report.
        **kwargs: Additional arguments for training.

        Prints:
        - A warning message for specified parameters that cannot be changed when that is attempted.
        """
        train_dataset = self.prepare_data(trainX, trainY)
        self.train(train_dataset, **kwargs)
        test_dataset = self.prepare_data(testX, testY)
        self.evaluate(test_dataset, mcm, classification_report)
        self.model.save_pretrained(self.output_path)

    def find_labels(self, trainY):
        """
        Finds and sets the unique labels in the training data.

        Parameters:
        trainY (list): List of training labels.
        """
        self.mlb = MultiLabelBinarizer()
        Y = [[i] for i in trainY]
        self.mlb.fit(Y)
        self.labels = [i for i in self.mlb.classes_]
        self.num_labels = len(self.labels)
    
    def prepare_data(self, X, Y):
        """
        Prepares training and test datasets for Transformers model training.

        Parameters:
        X (list): List of data (mentions).
        Y (list): List of labels.
        
        Returns:
        TensorDataset: TensorDataset containing tokenized data and labels.
        """
        tokenized_data = self.tokenizer(X, truncation=True, padding=True, return_tensors="pt", max_length=512)
        label_strings = [[i] for i in Y]
        labels = self.mlb.transform(label_strings)
        labels = torch.tensor(labels, dtype=torch.float32)
        tensordataset = TensorDataset(tokenized_data.input_ids, tokenized_data.attention_mask, labels)
        return tensordataset

    def train(self, train_dataset, **kwargs):
        """
        Trains the Transformers model on the provided dataset.

        Parameters:
        train_dataset (TensorDataset): Dataset for training.
        **kwargs: Additional arguments for training.

        Prints:
        - A warning message for specified parameters that cannot be changed.
        """
        training_args = TrainingArguments(
            output_dir="./output",  
            num_train_epochs=3,  
            per_device_train_batch_size=32, 
            evaluation_strategy="steps",
            save_steps=500,
            save_total_limit=2,
            load_best_model_at_end=True,
        )
        specified_params = ['data_collator', 'args','model']
        for param in specified_params:
            if param in kwargs:
                del kwargs[param]
                print("Warning. The parameter " + param + " cannot be changed from its given value.")
        self.trainer = Trainer(
            model=self.model,
            args=training_args,
            data_collator=self.collate_fn, 
            train_dataset=train_dataset,
            **kwargs,
        )
        self.trainer.train()

    def evaluate(self, testset, mcm, classification_report):
        """
        Evaluates the Transformers model on the provided test dataset.

        Parameters:
        testset (TensorDataset): Test dataset.
        mcm (bool): Whether to create a multi-label confusion matrix heatmap.
        classification_report (bool): Whether to display a classification report.

        Prints:
        - Evaluation metrics based on provided options.
        """
        results = self.trainer.predict(testset)
        max_indices = np.argmax(results.predictions, axis=1)
        preds = np.zeros_like(results.predictions)
        preds[np.arange(len(max_indices)), max_indices] = 1
        metrics = self.evaluate_model(preds, results.label_ids, mcm, classification_report)
        print(metrics)
        
    def collate_fn(self, batch):
        """
        Collates batches for training.

        Parameters:
        batch (list): Batch data containing input_ids, attention_mask, and labels.

        Returns:
        dict: Dictionary containing the batched input_ids, attention_mask, and labels.
        """
        return {
        'input_ids': torch.stack([item[0] for item in batch]),
        'attention_mask': torch.stack([item[1] for item in batch]),
        'labels': torch.stack([item[2] for item in batch])
        }
